@@ -18,7 +18,7 @@
  * 02110-1301  USA
  */
 
-#include <config.h>
+#include "config.h"
 
 #include <ClientTest.h>
 
@@ -35,10 +35,13 @@
 # include <execinfo.h>
 #endif
 
-#include "EvolutionSyncClient.h"
+#include "CmdlineSyncClient.h"
 #include "EvolutionSyncSource.h"
-#include "SyncEvolutionUtil.h"
-#include "VolatileConfigNode.h"
+#include <syncevo/util.h>
+#include <syncevo/VolatileConfigNode.h>
+
+#include <syncevo/declarations.h>
+SE_BEGIN_CXX
 
 /*
  * always provide this test class, even if not used:
@@ -47,73 +50,6 @@
  */
 CPPUNIT_REGISTRY_ADD_TO_DEFAULT("SyncEvolution");
 
-/**
- * a wrapper class which automatically does an open() in the constructor and a close() in the destructor
- * and ensures that the sync mode is "none" = testing mode
- */
-class TestEvolutionSyncSource : public EvolutionSyncSource {
-public:
-    TestEvolutionSyncSource(const string &type,
-                            const EvolutionSyncSourceParams &params) :
-        EvolutionSyncSource(params)
-    {
-        PersistentEvolutionSyncSourceConfig config(params.m_name, params.m_nodes);
-        config.setSourceType(type);
-        m_source.reset(EvolutionSyncSource::createSource(params));
-        CPPUNIT_ASSERT(m_source.get());
-    }
-
-    virtual SyncMLStatus beginSync(SyncMode mode) throw () {
-        CPPUNIT_ASSERT_NO_THROW(m_source->open());
-        CPPUNIT_ASSERT(!m_source->hasFailed());
-        return m_source->beginSync(mode);
-    }
-
-    virtual SyncMLStatus endSync() throw () {
-        SyncMLStatus res = m_source->endSync();
-        CPPUNIT_ASSERT_NO_THROW(m_source->close());
-        CPPUNIT_ASSERT(!m_source->hasFailed());
-        return res;
-    }
-
-    virtual SyncItem* getFirstItem() throw () { return m_source->getFirstItem(); }
-    virtual SyncItem* getNextItem() throw () { return m_source->getNextItem(); }
-    virtual SyncItem* getFirstNewItem() throw () { return m_source->getFirstNewItem(); }
-    virtual SyncItem* getNextNewItem() throw () { return m_source->getNextNewItem(); }
-    virtual SyncItem* getFirstUpdatedItem() throw () { return m_source->getFirstUpdatedItem(); }
-    virtual SyncItem* getNextUpdatedItem() throw () { return m_source->getNextUpdatedItem(); }
-    virtual SyncItem* getFirstDeletedItem() throw () { return m_source->getFirstDeletedItem(); }
-    virtual SyncItem* getNextDeletedItem() throw () { return m_source->getNextDeletedItem(); }
-
-    virtual SyncMLStatus addItem(SyncItem& item) throw () { return m_source->addItem(item); }
-    virtual SyncMLStatus updateItem(SyncItem& item) throw () { return m_source->updateItem(item); }
-    virtual SyncMLStatus deleteItem(SyncItem& item) throw () { return m_source->deleteItem(item); }
-    virtual SyncMLStatus removeAllItems() throw () { return m_source->removeAllItems(); }
-    const char *getName() throw () { return m_source->getName(); }
-
-    virtual Databases getDatabases() { return m_source->getDatabases(); }
-    virtual void open() { m_source->open(); }
-    virtual SyncItem *createItem(const string &uid, const char *type) { return m_source->createItem(uid, type); }
-    virtual void close() { m_source->close(); }
-    virtual void backupData(const string &dir, ConfigNode &node, BackupReport &report) { m_source->backupData(dir, node, report); }
-    virtual void restoreData(const string &dir, const ConfigNode &node, bool dryrun, SyncSourceReport &report) { m_source->restoreData(dir, node, dryrun, report); }
-    virtual const char *getMimeType() const { return m_source->getMimeType(); }
-    virtual const char *getMimeVersion() const { return m_source->getMimeVersion(); }
-    virtual const char* getSupportedTypes() const { return m_source->getSupportedTypes(); }
-    virtual bool checkStatus() { return m_source->checkStatus(); }
-    virtual void beginSyncThrow(bool needAll,
-                                bool needPartial,
-                                bool deleteLocal) { m_source->beginSyncThrow(needAll, needPartial, deleteLocal); }
-    virtual void endSyncThrow() { m_source->endSyncThrow(); }
-    virtual SyncMLStatus addItemThrow(SyncItem& item) { return m_source->addItemThrow(item); }
-    virtual SyncMLStatus updateItemThrow(SyncItem& item) { return m_source->updateItemThrow(item); }
-    virtual SyncMLStatus deleteItemThrow(SyncItem& item) { return m_source->deleteItemThrow(item); }
-    virtual void logItem(const string &uid, const string &info, bool debug = false) { m_source->logItem(uid, info, debug); }
-    virtual void logItem(const SyncItem &item, const string &info, bool debug = false) { m_source->logItem(item, info, debug); }
-
-    auto_ptr<EvolutionSyncSource> m_source;
-};
-
 class EvolutionLocalTests : public LocalTests {
 public:
     EvolutionLocalTests(const std::string &name, ClientTest &cl, int sourceParam, ClientTest::Config &co) :
@@ -154,16 +90,34 @@ private:
         update(createSourceA, item.c_str(), false);
 
         // opening and preparing the source should delete the item
-        std::auto_ptr<SyncSource> source;
+        std::auto_ptr<TestingSyncSource> source;
         SOURCE_ASSERT_NO_FAILURE(source.get(), source.reset(createSourceA()));
-        SOURCE_ASSERT(source.get(), source->beginSync(SYNC_NONE) == 0 );
-        CPPUNIT_ASSERT_EQUAL(0, countItemsOfType(source.get(), TOTAL_ITEMS));
-        CPPUNIT_ASSERT_EQUAL(0, countItemsOfType(source.get(), NEW_ITEMS));
-        CPPUNIT_ASSERT_EQUAL(0, countItemsOfType(source.get(), UPDATED_ITEMS));
-        CPPUNIT_ASSERT_EQUAL(1, countItemsOfType(source.get(), DELETED_ITEMS));
+        SOURCE_ASSERT_NO_FAILURE(source.get(), source->open());
+        SOURCE_ASSERT_NO_FAILURE(source.get(), source->beginSync("", "") );
+        CPPUNIT_ASSERT_EQUAL(0, countItemsOfType(source.get(), SyncSourceChanges::ANY));
+        CPPUNIT_ASSERT_EQUAL(0, countItemsOfType(source.get(), SyncSourceChanges::NEW));
+        CPPUNIT_ASSERT_EQUAL(0, countItemsOfType(source.get(), SyncSourceChanges::UPDATED));
+        CPPUNIT_ASSERT_EQUAL(1, countItemsOfType(source.get(), SyncSourceChanges::DELETED));
     }
 };
 
+/**
+ * This is a workaround for libecal/libebook in Evolution >= 2.30.
+ * The storage daemons shut down after only 10 seconds of no client
+ * being attached.  Due to limitations in libecal/libebook this is not
+ * detected when only using the synchronous API ("destroyed" signal
+ * not delivered, see e-cal.c), which then leads to D-Bus errors.
+ *
+ * The workaround consists of keeping one open SyncEvolution backend
+ * around for each of ical20 and vcard21/30, if they ever were used
+ * during testing.
+ */
+static map<string, boost::shared_ptr<TestingSyncSource> > lockEvolution;
+static void CleanupSources()
+{
+    lockEvolution.clear();
+}
+
 /**
  * This code uses the ClientTest and and information provided by
  * the backends in their RegisterSyncSourceTest instances to test
@@ -248,40 +202,57 @@ public:
 
         BOOST_FOREACH(const RegisterSyncSourceTest *test, m_configs) {
             if (sources.find(test->m_configName) != sources.end()) {
-                m_source2Config.push_back(test->m_configName);
+                m_syncSource2Config.push_back(test->m_configName);
             }
         }
 
+        /* Local Test SyncSource : remove all virtual datastores, inserting the
+         * sub datastores*/
+        ClientTest::Config conf;
+        BOOST_FOREACH (string source, sources) {
+            getSourceConfig (source, conf);
+            if (conf.subConfigs) {
+                vector<string> subs;
+                boost::split (subs, conf.subConfigs, boost::is_any_of(","));
+                BOOST_FOREACH (string sub, subs) {
+                    pushLocalSource2Config(sub);
+                }
+            } else {
+                pushLocalSource2Config(source);
+            }
+        }
         // get configuration and set obligatory fields
         LoggerBase::instance().setLevel(Logger::DEBUG);
         std::string root = std::string("evolution/") + server + "_" + id;
-        EvolutionSyncConfig config(string(server) + "_" + id);
-        boost::shared_ptr<EvolutionSyncConfig> from = boost::shared_ptr<EvolutionSyncConfig> ();
+        boost::shared_ptr<SyncConfig> config(new SyncConfig(string(server) + "_" + id));
+        boost::shared_ptr<SyncConfig> from = boost::shared_ptr<SyncConfig> ();
 
-        if (!config.exists()) {
-            // no configuration yet
-            config.setDefaults();
-            from = EvolutionSyncConfig::createServerTemplate(server);
+        if (!config->exists()) {
+            // no configuration yet, create in different contexts because
+            // device ID is different
+            config.reset(new SyncConfig(string(server) + "_" + id + "@client-test-" + id));
+            config->setDefaults();
+            from = SyncConfig::createPeerTemplate(server);
             if(from) {
                 set<string> filter;
-                config.copy(*from, &filter);
+                config->copy(*from, &filter);
             }
-            config.setDevID(id == "1" ? "sc-api-nat" : "sc-pim-ppc");
+            config->setDevID(id == "1" ? "sc-api-nat" : "sc-pim-ppc");
         }
         BOOST_FOREACH(const RegisterSyncSourceTest *test, m_configs) {
             ClientTest::Config testconfig;
             getSourceConfig(test, testconfig);
             CPPUNIT_ASSERT(testconfig.type);
 
-            boost::shared_ptr<EvolutionSyncSourceConfig> sc = config.getSyncSourceConfig(testconfig.sourceName);
+            boost::shared_ptr<SyncSourceConfig> sc = config->getSyncSourceConfig(testconfig.sourceName);
             if (!sc || !sc->exists()) {
                 // no configuration yet
-                config.setSourceDefaults(testconfig.sourceName);
-                sc = config.getSyncSourceConfig(testconfig.sourceName);
+                config->setSourceDefaults(testconfig.sourceName);
+                sc = config->getSyncSourceConfig(testconfig.sourceName);
                 CPPUNIT_ASSERT(sc);
                 sc->setURI(testconfig.uri);
                 if(from && testconfig.sourceNameServerTemplate){
-                    boost::shared_ptr<EvolutionSyncSourceConfig> scServerTemplate = from->getSyncSourceConfig(testconfig.sourceNameServerTemplate);
+                    boost::shared_ptr<SyncSourceConfig> scServerTemplate = from->getSyncSourceConfig(testconfig.sourceNameServerTemplate);
                     sc->setURI(scServerTemplate->getURI());
                 }
                 sc->setSourceType(testconfig.type);
@@ -293,19 +264,41 @@ public:
             sc->setUser(m_evoUser);
             sc->setPassword(m_evoPassword);
         }
-        config.flush();
+        config->flush();
     }
 
     virtual LocalTests *createLocalTests(const std::string &name, int sourceParam, ClientTest::Config &co) {
         return new EvolutionLocalTests(name, *this, sourceParam, co);
     }
 
-    virtual int getNumSources() {
-        return m_source2Config.size();
+    virtual int getNumLocalSources() {
+        return m_localSource2Config.size();
     }
 
-    virtual void getSourceConfig(int source, Config &config) {
-        getSourceConfig(m_configs[m_source2Config[source]], config);
+    virtual int getNumSyncSources() {
+        return m_syncSource2Config.size();
+    }
+
+    virtual void getLocalSourceConfig(int source, Config &config) {
+        getSourceConfig(m_configs[m_localSource2Config[source]], config);
+    }
+
+    virtual void getSyncSourceConfig(int source, Config &config) {
+        getSourceConfig(m_configs[m_syncSource2Config[source]], config);
+    }
+
+    virtual int getLocalSourcePosition(const string &configName) {
+        for (size_t i=0; i< m_localSource2Config.size(); i++) {
+            if(m_localSource2Config[i] == configName) {
+                return i;
+                break;
+            }
+        }
+        return -1;
+    }
+
+    virtual void getSourceConfig (const string &configName, Config &config) {
+        return getSourceConfig (m_configs[configName], config);
     }
 
     static void getSourceConfig(const RegisterSyncSourceTest *test, Config &config) {
@@ -330,22 +323,16 @@ public:
                                 const std::string &logbase,
                                 const SyncOptions &options)
     {
-        set<string> activeSources;
-        for(int i = 0; sources[i] >= 0; i++) {
-            activeSources.insert(m_source2Config[sources[i]]);
-        }
-
         string server = getenv("CLIENT_TEST_SERVER") ? getenv("CLIENT_TEST_SERVER") : "funambol";
         server += "_";
         server += m_clientID;
         
-        class ClientTest : public EvolutionSyncClient {
+        class ClientTest : public CmdlineSyncClient {
         public:
             ClientTest(const string &server,
-                       const set<string> &activeSources,
                        const string &logbase,
                        const SyncOptions &options) :
-                EvolutionSyncClient(server, false, activeSources),
+                CmdlineSyncClient(server, false, true),
                 m_logbase(logbase),
                 m_options(options),
                 m_started(false)
@@ -358,12 +345,9 @@ public:
                 setMaxObjSize(m_options.m_maxObjSize, true);
                 setMaxMsgSize(m_options.m_maxMsgSize, true);
                 setWBXML(m_options.m_isWBXML, true);
-                EvolutionSyncClient::prepare();
-            }
-            virtual void prepare(const std::vector<EvolutionSyncSource *> &sources) {
-                SyncModes modes(m_options.m_syncMode);
-                setSyncModes(sources, modes);
-                EvolutionSyncClient::prepare(sources);
+                setRetryDuration(m_options.m_retryDuration, true);
+                setRetryInterval(m_options.m_retryInterval, true);
+                SyncContext::prepare();
             }
 
             virtual void displaySyncProgress(sysync::TProgressEventEnum type,
@@ -383,7 +367,7 @@ public:
             virtual boost::shared_ptr<TransportAgent> createTransportAgent()
             {
                 boost::shared_ptr<TransportAgent>wrapper = m_options.m_transport;
-                boost::shared_ptr<TransportAgent>agent =EvolutionSyncClient::createTransportAgent();
+                boost::shared_ptr<TransportAgent>agent =SyncContext::createTransportAgent();
                 if (!wrapper.get())
                     return agent;
                 dynamic_cast<TransportWrapper*>(wrapper.get())->setAgent(agent);
@@ -395,7 +379,20 @@ public:
             const string m_logbase;
             SyncOptions m_options;
             bool m_started;
-        } client(server, activeSources, logbase, options);
+        } client(server, logbase, options);
+
+        // configure active sources with the desired sync mode,
+        // disable the rest
+        FilterConfigNode::ConfigFilter filter;
+        filter[SyncSourceConfig::m_sourcePropSync.getName()] = "none";
+        client.setConfigFilter(false, "", filter);
+        filter[SyncSourceConfig::m_sourcePropSync.getName()] =
+            PrettyPrintSyncMode(options.m_syncMode);
+        for(int i = 0; sources[i] >= 0; i++) {
+            std::string &name = m_syncSource2Config[sources[i]];
+            client.setConfigFilter(false, name, filter);
+            checkEvolutionSource(name);
+        }
 
         SyncReport report;
         SyncMLStatus status = client.sync(&report);
@@ -416,40 +413,94 @@ private:
      * This is the mapping to the corresponding config name, created when
      * constructing this instance.
      */
-    vector<string> m_source2Config;
+    vector<string> m_localSource2Config;
+    vector<string> m_syncSource2Config;
 
     /** returns the name of the Evolution database */
     string getDatabaseName(const string &configName) {
+        if (configName == "calendar+todo") {
+            return "ical20,itodo20";
+        } else if (configName == "file_calendar+todo") {
+            return "file_ical20,file_itodo20";
+        }
         return m_evoPrefix + configName + "_" + m_clientID;
     }
-    
-    static SyncSource *createSource(ClientTest &client, int source, bool isSourceA) {
-        TestEvolution &evClient((TestEvolution &)client);
-        string changeID = "SyncEvolution Change ID #";
-        string name = evClient.m_source2Config[source];
-        changeID += isSourceA ? "1" : "2";
-        string database = evClient.getDatabaseName(name);
 
-        EvolutionSyncConfig config("client-test-changes");
+    /** called by test frame work */
+    static TestingSyncSource *createSource(ClientTest &client, int source, bool isSourceA) {
+        TestEvolution &evClient((TestEvolution &)client);
+        string name = evClient.m_localSource2Config[source];
+
+        // implement Evolution shutdown workaround (see lockEvolution above)
+        evClient.checkEvolutionSource(name);
+
+        return evClient.createSource(name, isSourceA);
+    }
+
+    /** called internally in this class */
+    TestingSyncSource *createSource(const string &name, bool isSourceA) {
+        string database = getDatabaseName(name);
+        SyncConfig config("client-test-changes");
         SyncSourceNodes nodes = config.getSyncSourceNodes(name,
-                                                          string("_") + ((TestEvolution &)client).m_clientID +
+                                                          string("_") + m_clientID +
                                                           "_" + (isSourceA ? "A" : "B"));
 
         // always set this property: the name might have changes since last test run
-        nodes.m_configNode->setProperty("evolutionsource", database.c_str());
-        nodes.m_configNode->setProperty("evolutionuser", evClient.m_evoUser.c_str());
-        nodes.m_configNode->setProperty("evolutionpassword", evClient.m_evoPassword.c_str());
+        nodes.getProperties()->setProperty("evolutionsource", database.c_str());
+        nodes.getProperties()->setProperty("evolutionuser", m_evoUser.c_str());
+        nodes.getProperties()->setProperty("evolutionpassword", m_evoPassword.c_str());
 
-        EvolutionSyncSourceParams params(name,
-                                         nodes,
-                                         changeID);
+        SyncSourceParams params(name,
+                                nodes);
 
-        const RegisterSyncSourceTest *test = evClient.m_configs[name];
+        const RegisterSyncSourceTest *test = m_configs[name];
         ClientTestConfig testConfig;
         getSourceConfig(test, testConfig);
 
-        SyncSource *ss = new TestEvolutionSyncSource(testConfig.type, params);
-        return ss;
+        PersistentSyncSourceConfig sourceConfig(params.m_name, params.m_nodes);
+        sourceConfig.setSourceType(testConfig.type);
+
+        // downcasting here: anyone who registers his sources for testing
+        // must ensure that they are indeed TestingSyncSource instances
+        SyncSource *ss = SyncSource::createSource(params);
+        return static_cast<TestingSyncSource *>(ss);
+    }
+
+    // push source into localsource2config if it doesn't exist in the vector
+    void pushLocalSource2Config(const string &source) {
+        bool finded = false;
+        BOOST_FOREACH(string element, m_localSource2Config) {
+            if (boost::iequals(element, source)) {
+                finded = true;
+                break;
+            }
+        }
+        if (!finded) {
+            m_localSource2Config.push_back (source);
+        }
+    }
+
+    void checkEvolutionSource(std::string &name)
+    {
+        string basename;
+
+        // hard-coded names as used by src/backends/evolution;
+        // if some other backend reuses them, it gets the
+        // same treatment, which shouldn't cause any harm
+        if (name == "vcard21" ||
+            name == "vcard30") {
+            basename = "ebook";
+        } else if (name == "ical20" ||
+                   name == "text") {
+            basename = "ecal";
+        }
+
+        if (!basename.empty() &&
+            lockEvolution.find(basename) == lockEvolution.end()) {
+            lockEvolution[basename].reset(createSource(name, true));
+            lockEvolution[basename]->open();
+            ClientTest::registerCleanup(CleanupSources);
+        }
     }
 };
 
@@ -491,6 +542,7 @@ public:
         // this is required when using glib directly or indirectly
         g_type_init();
         g_thread_init(NULL);
+        g_set_prgname("client-test");
 #endif
         EDSAbiWrapperInit();
         testClient.registerTests();
@@ -500,3 +552,5 @@ private:
     TestEvolution testClient;
     
 } testEvolution;
+
+SE_END_CXX
